#pragma once
#include "util/i_sleeper.h"

#include <atomic>
#include <chrono>
#include <condition_variable>
#include <iostream>
#include <mutex>
#include <thread>

namespace spin
{
    class Sleeper : public ISleeper
    {
    public:
        ~Sleeper()
        {
                wakeUpAll();
        }

        void sleepFor(const std::chrono::milliseconds &duration) override
        {
            startWakeUpTask();
            std::unique_lock<std::mutex> lock(mutex_);
            cv_.wait_for(lock, duration);
        }

        void sleepUntil(const std::chrono::time_point<std::chrono::system_clock> &timePoint) override
        {
            startWakeUpTask();
            std::unique_lock<std::mutex> lock(mutex_);
            cv_.wait_until(lock, timePoint);
        }

        void startWakeUpTask()
        {
            if (keepSleeping_.load())
            {
                return;
            }
            keepSleeping_.store(true);
            wakeUpTask_ = std::thread([this](){
                while(keepSleeping_.load());
                std::cout << "Wake up!\n";
                std::unique_lock<std::mutex> lock(mutex_);
                cv_.notify_all();
            });
        }

        void wakeUpAll() override
        {
            keepSleeping_.store(false);
            if (wakeUpTask_.joinable() && wakeUpTask_.get_id() != std::this_thread::get_id())
            {
                wakeUpTask_.join();
            }
        }

    private:
        std::mutex mutex_;
        std::condition_variable cv_;
        std::thread wakeUpTask_;
        std::atomic<bool> keepSleeping_{false};
    };
    
} // namespace spin